// David Eberly, Geometric Tools, Redmond WA 98052
// Copyright (c) 1998-2016
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
// File Version: 3.0.1 (2016/06/29)

#pragma once

#include <LowLevel/GteLogger.h>
#include <Mathematics/GteDarbouxFrame.h>
#include <memory>

namespace gte
{

template <typename Real>
class RectangleSurface
{
public:
    // Construction.  This class is designed to support triangulations of
    // surfaces where the vertices of the triangle have positions and
    // optionally normals, tangents, bitangents, and texture coordinates.
    // The vertices must be stored in an array where the stride is the
    // number of bytes between the addresses of consecutive vertices; that
    // is, stride = &vertex[1] - &vertex[0].  The 'surface' must be
    // rectangular; that is, surface->IsRectangular() must return true, and
    // the stride must be positive.  All vertex attributes must have Real
    // components.
    RectangleSurface(std::shared_ptr<ParametricSurface<3, Real>> const& surface, size_t stride);

    // Member access.
    inline std::shared_ptr<ParametricSurface<3, Real>> const& GetSurface() const;
    inline size_t GetStride() const;

    // Support for generating triangle meshes that approximate the rectangle
    // surface.  The numUSamples and numVSamples must be 2 or larger.  An
    // example of use:
    //
    //   struct Vertex
    //   {
    //       Vector4<float> position;
    //       Vector4<float> normal;
    //       Vector2<float> tcood;
    //   };
    //
    //   RectangleSurface rect(surface, stride);
    //   rect.Set(numUSamples, numVSamples);
    //
    //   unsigned int numVertices = rect.GetNumVertices();
    //   std::vector<Vertex> vertices(numVertices);
    //
    //   unsigned int numTriangles = rect.GetNumTriangles();
    //   std::vector<unsigned int> indices(3 * numTriangles);
    //   rect.GetIndices(indices.data());
    //
    //   // Tessellate the surface but compute only the position and tcoord
    //   // attributes of the mesh.
    //   rect.Tessellate(&vertices[0].position, nullptr, nullptr, nullptr, &vertices[0].tcoord);
    //
    // If the underlying surface varies over time, you can call the Tessellate
    // function at each time of interest to obtain a time-varying mesh.
    void Set(unsigned int numUSamples, unsigned int numVSamples);
    inline unsigned int GetNumUSamples() const;
    inline unsigned int GetNumVSamples() const;
    inline unsigned int GetNumVertices() const;
    inline unsigned int GetNumTriangles() const;
    void GetIndices(unsigned int* indices) const;

    // The positions, normals, tangents, and bitangents must be 3-tuples
    // (x,y,z) or 4-tuples (x,y,z,w), where w = 1 for positions and w = 0
    // for normals, tangents, and bitangents.  In the 4-tuple case, the
    // evaluations modify only the (x,y,z) portion of the attribute.  The
    // texture coordinates must be 2-tuples (x,y).  The caller is responsible
    // for providing the correct pointers to the vertex attributes.  If you
    // want an attribute evaluated, pass a nonnull pointer.  If a pointer is
    // null, that attribute is not processed.
    void Tessellate(void* positions, void* normals, void* tangents, void* bitangents, void* tcoords);

private:
    inline Vector3<Real>& Position(unsigned int i);
    inline Vector3<Real>& Normal(unsigned int i);
    inline Vector3<Real>& Tangent(unsigned int i);
    inline Vector3<Real>& Bitangent(unsigned int i);
    inline Vector2<Real>& TCoord(unsigned int i);

    std::shared_ptr<ParametricSurface<3, Real>> mSurface;
    size_t mStride;
    unsigned int mNumUSamples, mNumVSamples;
    unsigned int mNumVertices, mNumTriangles;
    char* mPositions;
    char* mNormals;
    char* mTangents;
    char* mBitangents;
    char* mTCoords;
    Vector3<Real> mInvalid3;
    Vector2<Real> mInvalid2;
};


template <typename Real>
RectangleSurface<Real>::RectangleSurface(std::shared_ptr<ParametricSurface<3, Real>> const& surface,
    size_t stride)
    :
    mSurface(surface),
    mStride(stride),
    mNumUSamples(0),
    mNumVSamples(0),
    mNumVertices(0),
    mNumTriangles(0),
    mPositions(nullptr),
    mNormals(nullptr),
    mTangents(nullptr),
    mBitangents(nullptr),
    mTCoords(nullptr),
    mInvalid3({ (Real)0, (Real)0, (Real)0 }),
    mInvalid2({ (Real)0, (Real)0 })
{
    if (!mSurface || !mSurface->IsRectangular() || mStride == 0)
    {
        LogError("Invalid surface.");
        mSurface = nullptr;
        mStride = 0;
    }
}

template <typename Real> inline
std::shared_ptr<ParametricSurface<3, Real>> const& RectangleSurface<Real>::GetSurface() const
{
    return mSurface;
}

template <typename Real>inline
size_t RectangleSurface<Real>::GetStride() const
{
    return mStride;
}

template <typename Real>
void RectangleSurface<Real>::Set(unsigned int numUSamples, unsigned int numVSamples)
{
    if (numUSamples >= 2 && numVSamples >= 2)
    {
        mNumUSamples = numUSamples;
        mNumVSamples = numVSamples;
        mNumVertices = numUSamples * numVSamples;
        mNumTriangles = 2 * (numUSamples - 1) * (numVSamples - 1);
    }
    else
    {
        LogError("Invalid number of samples.");
        mSurface = nullptr;
        mStride = 0;
    }
}

template <typename Real>inline
unsigned int RectangleSurface<Real>::GetNumUSamples() const
{
    return mNumUSamples;
}

template <typename Real>inline
unsigned int RectangleSurface<Real>::GetNumVSamples() const
{
    return mNumVSamples;
}

template <typename Real> inline
unsigned int RectangleSurface<Real>::GetNumVertices() const
{
    return mNumVertices;
}

template <typename Real> inline
unsigned int RectangleSurface<Real>::GetNumTriangles() const
{
    return mNumTriangles;
}

template <typename Real>
void RectangleSurface<Real>::GetIndices(unsigned int* indices) const
{
    if (mSurface)
    {
        for (unsigned int v = 0, i = 0; v + 1 < mNumVSamples; ++v)
        {
            unsigned int i0 = i;
            unsigned int i1 = i0 + 1;
            i += mNumUSamples;
            unsigned int i2 = i;
            unsigned int i3 = i2 + 1;
            for (unsigned int u = 0; u + 1 < mNumUSamples; ++u, indices += 6)
            {
                indices[0] = i0;
                indices[1] = i2;
                indices[2] = i1;
                indices[3] = i1;
                indices[4] = i2;
                indices[5] = i3;
                ++i0;
                ++i1;
                ++i2;
                ++i3;
            }
        }
    }
}

template <typename Real>
void RectangleSurface<Real>::Tessellate(void* positions, void* normals,
    void* tangents, void* bitangents, void* tcoords)
{
    if (!mSurface)
    {
        return;
    }

    mPositions = static_cast<char*>(positions);
    mNormals = static_cast<char*>(normals);
    mTangents = static_cast<char*>(tangents);
    mBitangents = static_cast<char*>(bitangents);
    mTCoords = static_cast<char*>(tcoords);

    Real uMin = mSurface->GetUMin();
    Real uDelta = (mSurface->GetUMax() - uMin) / static_cast<Real>(mNumUSamples - 1);
    Real vMin = mSurface->GetVMin();
    Real vDelta = (mSurface->GetVMax() - vMin) / static_cast<Real>(mNumVSamples - 1);
    DarbouxFrame3<Real> darboux(mSurface);
    for (unsigned int vIndex = 0, i = 0; vIndex < mNumVSamples; ++vIndex)
    {
        Real v = vMin + vDelta * vIndex;
        for (unsigned int uIndex = 0; uIndex < mNumUSamples; ++uIndex, ++i)
        {
            Real u = uMin + uDelta * uIndex;

            if (mNormals || mTangents || mBitangents)
            {
                darboux(u, v, Position(i), Tangent(i), Bitangent(i), Normal(i));
            }
            else
            {
                Position(i) = mSurface->GetPosition(u, v);
            }

            TCoord(i) = { u, v };
        }
    }

    mPositions = nullptr;
    mNormals = nullptr;
    mTangents = nullptr;
    mBitangents = nullptr;
    mTCoords = nullptr;
}

template <typename Real> inline
Vector3<Real>& RectangleSurface<Real>::Position(unsigned int i)
{
    if (mPositions)
    {
        return *reinterpret_cast<Vector3<Real>*>(mPositions + i * mStride);
    }
    else
    {
        return mInvalid3;
    }
}

template <typename Real> inline
Vector3<Real>& RectangleSurface<Real>::Normal(unsigned int i)
{
    if (mNormals)
    {
        return *reinterpret_cast<Vector3<Real>*>(mNormals + i * mStride);
    }
    else
    {
        return mInvalid3;
    }
}

template <typename Real> inline
Vector3<Real>& RectangleSurface<Real>::Tangent(unsigned int i)
{
    if (mTangents)
    {
        return *reinterpret_cast<Vector3<Real>*>(mTangents + i * mStride);
    }
    else
    {
        return mInvalid3;
    }
}

template <typename Real> inline
Vector3<Real>& RectangleSurface<Real>::Bitangent(unsigned int i)
{
    if (mBitangents)
    {
        return *reinterpret_cast<Vector3<Real>*>(mBitangents + i * mStride);
    }
    else
    {
        return mInvalid3;
    }
}

template <typename Real> inline
Vector2<Real>& RectangleSurface<Real>::TCoord(unsigned int i)
{
    if (mTCoords)
    {
        return *reinterpret_cast<Vector2<Real>*>(mTCoords + i * mStride);
    }
    else
    {
        return mInvalid2;
    }
}

}
